import json
from pathlib import Path

from django.conf import settings

from rest_framework import status as drf_status

from rest_framework.test import APIClient
from rest_framework_simplejwt.tokens import RefreshToken
from rest_framework.authtoken.models import Token

from w import Debug
from w.django.fixtures.factory_boy import UserFactory
from w.django.tests.django_testcase import DjangoTestCase
from w.django.utils import reverse


class ApiTestCase(DjangoTestCase):
    """ Class to handle common API test functionalities """

    resource = None
    exclude_key_check = []
    show_json_response = False
    default_user = None
    default_token = None
    client_class = APIClient

    @classmethod
    def setUpTestData(cls):
        cls.factory_boy_seed_random()

    @staticmethod
    def get_resource_objects():
        raise RuntimeError("To be implemented into child class")

    @classmethod
    def serialize_resource(cls, resource, many=False):
        return getattr(cls, "serialize_%s" % cls.resource)(resource, many)

    def _get_user(self):
        if self.default_user is None:
            self.default_user = UserFactory(
                username="factory-boy",
                first_name="Factory",
                last_name="Boy",
                email="fb@factory.com",
            )
        return self.default_user

    def force_authenticate(self, user=None):
        """
        Force authenticate, if no user given, use a default one
        (created via factory_boy)

        Args:
            user (User): user to authenticate (optional)
        """
        if user is None:
            user = self._get_user()
        self.client.force_authenticate(user=user)

    @staticmethod
    def _get_jwt_access_token(user):
        jwt_token = RefreshToken.for_user(user)
        return str(jwt_token.access_token)

    def _client_method(self, method, url, data=None, auth_by_token=True, user=None):
        params = {"path": url, "data": data}
        if auth_by_token:
            if user:
                token, created = Token.objects.get_or_create(user=user)
                params["HTTP_AUTHORIZATION"] = f"Token {token}"
            elif self.default_token is None:
                self.default_token, created = Token.objects.get_or_create(
                    user=self._get_user()
                )
                params["HTTP_AUTHORIZATION"] = f"Token {self.default_token}"
            else:
                # do not delete, case of multiple client.get in a single test
                params["HTTP_AUTHORIZATION"] = f"Token {self.default_token}"

        return getattr(self.client, method)(**params)

    def client_get(self, url, auth_by_token=True, user=None):
        return self._client_method("get", url, auth_by_token=auth_by_token, user=user)

    def client_post(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "post", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def client_put(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "put", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def client_delete(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "delete", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def client_patch(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "patch", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def _assert_no_registered_user_return_401(self, method, urlconf, url_params):
        response = getattr(self.client, method)(reverse(urlconf, kwargs=url_params))
        self.assert_rest_status_code(response, drf_status.HTTP_401_UNAUTHORIZED)

    def _assert_unauthorized_user_return_403(
        self, method, url_name, url_params, user=None, data=None
    ):
        params = {"url": reverse(url_name, kwargs=url_params), "user": user}
        if data is not None:
            params["data"] = data
        response = getattr(self, f"client_{method}")(**params)
        self.assert_rest_status_code(response, drf_status.HTTP_403_FORBIDDEN)

    def _assert_unknown_resource_return_404(self, method, urlconf, url_params):
        response = getattr(self, f"client_{method}")(
            reverse(urlconf, kwargs=url_params)
        )
        self.assert_rest_status_code(response, drf_status.HTTP_404_NOT_FOUND)

    def assert_get_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot get if no user authenticated """
        self._assert_no_registered_user_return_401("get", urlconf, url_params)

    def assert_get_with_unauthorized_user_return_403(
        self, user, urlconf, url_params=None
    ):
        """ Ensure we cannot get if user is not authorized """
        self._assert_unauthorized_user_return_403("get", user, urlconf, url_params)

    def assert_get_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot get if resource is unknown """
        self._assert_unknown_resource_return_404("get", urlconf, url_params)

    def assert_post_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot post if no user authenticated """
        self._assert_no_registered_user_return_401("post", urlconf, url_params)

    def assert_post_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot post if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "post", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_post_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot post if resource is unknown """
        self._assert_unknown_resource_return_404("post", urlconf, url_params)

    def assert_put_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot put if no user authenticated """
        self._assert_no_registered_user_return_401("put", urlconf, url_params)

    def assert_put_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot put if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "put", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_put_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot put if resource is unknown """
        self._assert_unknown_resource_return_404("put", urlconf, url_params)

    def assert_patch_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot patch if no user authenticated """
        self._assert_no_registered_user_return_401("patch", urlconf, url_params)

    def assert_patch_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot patch if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "patch", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_patch_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot patch if resource is unknown """
        self._assert_unknown_resource_return_404("patch", urlconf, url_params)

    def assert_delete_with_no_registered_user_return_401(
        self, urlconf, url_params=None
    ):
        """ Ensure we cannot delete if no user authenticated """
        self._assert_no_registered_user_return_401("delete", urlconf, url_params)

    def assert_delete_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot delete if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "delete", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_delete_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot delete if resource is unknown """
        self._assert_unknown_resource_return_404("delete", urlconf, url_params)

    def assert_get_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_200_OK)

    def assert_post_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_201_CREATED)

    def assert_put_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_200_OK)

    def assert_patch_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_200_OK)

    def assert_delete_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_204_NO_CONTENT)

    @staticmethod
    def assert_rest_status_code(response, status):
        try:
            assert (
                status == response.status_code
            ), f"expected status {status} == {response.status_code} actual"
        except Exception as e:
            Debug.s(json.loads(response.content))
            raise e

    def assert_rest_response(self, response, status):
        self.assert_rest_status_code(response, status)
        if self.show_json_response:
            Debug.s(response.content)
        self.assert_equals_resultset(json.loads(response.content))

    def assert_invalid_data_response(self, response):
        self.assert_rest_response(
            response, status=drf_status.HTTP_422_UNPROCESSABLE_ENTITY
        )

    def _assert_crud_response(self, data, response, status, actual, **kwargs):
        self.assert_rest_status_code(response, status)
        if self.show_json_response:
            Debug.s(response.content)
        content = json.loads(response.content)
        for k, v in data.items():
            if k not in self.exclude_key_check:
                assert content[k] == v
        self.assert_equals_resultset(self.serialize_resource(actual), **kwargs)

    def save_as_cypress_fixtures(self, fixture_path, dataset):
        """
        Save dataset into cypress fixture repo
        """
        if settings.CYPRESS_FIXTURE_DIR is None:
            # ignore cypress fixture saving
            return None

        new_json = (
            json.dumps(
                self._convert_data(dataset),
                indent=4,
                separators=(",", ": "),
                default=str,
                # disable non-ASCII characters escape with \uXXXX sequences
                ensure_ascii=False,
            )
            + "\n"
        )  # for respecting file empty last line convention

        fixture_filename = Path(f"{settings.CYPRESS_FIXTURE_DIR}/{fixture_path}")
        if fixture_filename.exists():
            # check if we need to update it
            actual = fixture_filename.read_text()
            if actual == new_json:
                return None
        else:
            self._check_file_can_be_created(fixture_filename)

        # save dataset
        fixture_filename.write_text(new_json)
